//
//  SpectrumUnit.cpp
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/20/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#include "SpectrumUnit.h"

SpectrumUnit::SpectrumUnit(const string & sInputFileName, const string & sScanType, int iDefaultPrecursorChargeState,
                           const SettingInfo * pSettingInfo)
//sInputFileName: name of spectra file (no path information); sScanType: scan type
//iDefaultPrecursorChargeState: default charge state of precursor; pSettingInfo: point to setting information
{
    this->sRetentionTime = "NA";
    this->sInchi = "";
    this->sFileName = sInputFileName;
    this->sScanType = sScanType;
    if (this->sScanType == "PositiveIon")
    {
        this->iPrecursorCharge = iDefaultPrecursorChargeState;//(+1)*iDefaultPrecursorChargeState
        this->viIonFragmentMassWindows = pSettingInfo->viPositiveIonFragmentMassWindows;
    }
    else
    {
        this->iPrecursorCharge = iDefaultPrecursorChargeState * (-1);
        this->viIonFragmentMassWindows = pSettingInfo->viNegativeIonFragmentMassWindows;
    }
    this->pSettingInfo = pSettingInfo;
    this->dHalf_mass_tolerance_fragment_ions = pSettingInfo->dMassToleranceFragmentIons * 0.5;
}

SpectrumUnit::~SpectrumUnit()
{
    
}

double SpectrumUnit::mypnorm(double dMean, double dStandardDeviation, double dRandomVariable)
{
    double dZScore = 0, dProbability = 0;
    
    dZScore = (dRandomVariable - dMean) / dStandardDeviation;
    dProbability = 0.5 * erfc((-1) * dZScore/sqrt(2.0));
    
    return dProbability;
}

double SpectrumUnit::calLamda(double dErrorMZ)
//dErrorMZ: dMeasuredMZ - dCalculatedMZ; return Lamda value
{
    double dLamda;
    
    dLamda = (1 - mypnorm(0, dHalf_mass_tolerance_fragment_ions, fabs(dErrorMZ)))*2.0;
    
    return dLamda;
}

void SpectrumUnit::PreprocessPeaks()
{
    sort(vPeaks.begin(), vPeaks.end(), lessMZ);
    NormalizeIntensity();
}

void SpectrumUnit::NormalizeIntensity()
{
    vector<peakType>::iterator itvPeakType;
    double dMaxIntensities = 0;
    
    for (itvPeakType = vPeaks.begin(); itvPeakType != vPeaks.end(); ++itvPeakType)
        if (dMaxIntensities < itvPeakType->dIntensity )
            dMaxIntensities = itvPeakType->dIntensity;
    
    if (dMaxIntensities > 0)
    {
        for (itvPeakType = vPeaks.begin(); itvPeakType != vPeaks.end(); ++itvPeakType)
            itvPeakType->dIntensity = itvPeakType->dIntensity / dMaxIntensities;
    }
}

bool SpectrumUnit::lessMZ(peakType record_i, peakType record_j)
//record_i and record_j are two peaks
//return true if mz of record_i is less than that of record_j.
{
    return (record_i.dMZ < record_j.dMZ);
}

scoreInfo SpectrumUnit::scoreCompound(const compoundType *pCompound)
// pCompound: pointer referring to the compound we are scoringpIntactCompound
// return the score information of this compound against the spectrum
{
    size_t i, iPeakNum;
    peakFragmentInfo initPeakFragInfo = {};
    scoreInfo curScoreCompoundMatch;
    fragType * pIntactCompound;
    RWMol * pIntactRWMol;
    GML* pCurGraph;
    int iExplainedPeaks = 0;
    
    iPeakNum = vPeaks.size();
    initPeakFragInfo.dPeakScore = 0;
    for (i = 0; i < iPeakNum; i++)
        curScoreCompoundMatch.vPeakFragInfo.push_back(initPeakFragInfo);
    curScoreCompoundMatch.dScore = 0;
    curScoreCompoundMatch.dParentMassError = pCompound->dMass - this->dPrecursorMass;
    curScoreCompoundMatch.sCompoundIdentifier = pCompound->sIdentifier;
    curScoreCompoundMatch.sInchi = pCompound->sInchi;
    curScoreCompoundMatch.sCompoundName = pCompound->sName;
    curScoreCompoundMatch.sCompoundLinks = pCompound->sLinks;
    
    pIntactRWMol = new RWMol(*(pCompound->pRWMol));
    pIntactCompound = new fragType();
    pIntactCompound->dFragmentMass = pCompound->dMass;
    pIntactCompound->pFragmentRWMol = pIntactRWMol;
    pIntactCompound->molTopology = pCompound->molTopology;
    if (pSettingInfo->bGraphGrammar)
    {
        pCurGraph = Database::ROMol2GML(pIntactCompound->pFragmentRWMol);
        pIntactCompound->rgFragTopology.vpSubgraphs.push_back(pCurGraph);
        pIntactCompound->vbGrammarPattern = pCompound->vbGrammarPattern;
    }
    pIntactCompound->iFragmentationLevel = 0;
    pIntactCompound->bParentFragmentDetected = true;
    pIntactCompound->dBondWeight = 0;
    pIntactCompound->sSmiles = MolToSmiles(*(pIntactCompound->pFragmentRWMol));
    //processFragment(curScoreCompoundMatch, pIntactCompound, 1.0);//new version
    processFragment_original(curScoreCompoundMatch, pIntactCompound, 1.0);//midas_1.0//1.0 is the Parent PlausibilityScore
    for (i=0; i<iPeakNum; i++)//just for midas 1.0
        curScoreCompoundMatch.dScore += curScoreCompoundMatch.vPeakFragInfo.at(i).dPeakScore;
    for (i=0; i<iPeakNum; i++)
    {
        if (curScoreCompoundMatch.vPeakFragInfo.at(i).sSmiles != "")
            iExplainedPeaks += 1;
    }
    curScoreCompoundMatch.sExplainedPeaks = to_string(iExplainedPeaks)+">"+to_string(iPeakNum);    
    delete pIntactCompound->pFragmentRWMol;
    delete pIntactCompound;
    
    return curScoreCompoundMatch;
}

double SpectrumUnit::calculatePlausibilityScore(const fragType *pCurFragment, const double dParentPlausibilityScore)
//pCurFragment: pointer to the current fragment; dParentPlausibilityScore: plausibility score of the parent fragment
//return the Plausibility Score of the current fragment
{
    double dCurrentPlausibilityScore;
    
    if (pCurFragment->bParentFragmentDetected)
        dCurrentPlausibilityScore = pow(0.5, pCurFragment->dBondWeight) * dParentPlausibilityScore;
    else
        dCurrentPlausibilityScore = pow(0.1, pCurFragment->dBondWeight) * dParentPlausibilityScore;
    
    return dCurrentPlausibilityScore;
}

void SpectrumUnit::calculateAllPossibleMZ(vector<double> & vdPossibleMZ, const fragType *pCurFragment,
                                            int iCurrentIonCharge)
//vdPossibleMZ: all possible mz; pCurFragment: pointer to the current fragment;
//iCurrentIonCharge: current ion (or peak) charge
{
    size_t i;
    double dCalculatedMZ;
    
    vdPossibleMZ.clear();
    for (i=0; i<this->viIonFragmentMassWindows.size(); i++)
    {
        dCalculatedMZ = (this->viIonFragmentMassWindows.at(i) * this->pSettingInfo->dProtonMass
                         + pCurFragment->dFragmentMass)/ iCurrentIonCharge;
        vdPossibleMZ.push_back(dCalculatedMZ);
    }
}

double SpectrumUnit::calculatePeakFragmentMatch(const peakType &curPeak,double dCurrentPlausibilityScore,double dErrorMZ)
//dCurrentPlausibilityScore: palusibility score of the current fragment
//dErrorMZ: dMeasuredMZ - dCalculatedMZ; curPeak: current peak;
//return the score of this match (specific peak, specific fragment, specific proton offset)
{
    double dSubScore;
    
    dSubScore = dCurrentPlausibilityScore * curPeak.dIntensity * calLamda(dErrorMZ);
    
    return dSubScore;
}
/*
double SpectrumUnit::calculateScoreFragmentAllPeaks(const vector<double> &vdCalculatedMZ,
                                                    vector<double> &vdCurCalculatedMZ, const fragType *pCurFragment,
                                                    size_t &iPeakRecordID, size_t &iMassWindowID,
                                                    double dCurrentPlausibilityScore)
//vdCalculatedMZ: precalculated mz for all ion mass windows, if peak charge is NOT equal to this->iDefaultIonChargeState,
// it should be recalculated; vdCurCalculatedMZ: mz for all ion mass windows of the current fragment
//pCurFragment: pointer to the current fragment; iPeakRecordID: peak record id with the highest score;
//iMassWindowID: mass window id with the highest score;
//dCurrentPlausibilityScore: plausibility score of the current fragment; return the highest score
{
    double dHighestPeakScore = -1, dErrorMZ, dCurPeakScore;
    size_t i, j, iPeakNum;
    peakType curPeak;
    
    iPeakRecordID=-1;
    iMassWindowID=-1;
    
    iPeakNum = this->vPeaks.size();
    for (i = 0; i < iPeakNum; i++)
    {
        curPeak = this->vPeaks.at(i);
        if (curPeak.iPeakCharge != this->iDefaultIonChargeState)
            calculateAllPossibleMZ(vdCurCalculatedMZ, pCurFragment, curPeak.iPeakCharge);
        else
            vdCurCalculatedMZ = vdCalculatedMZ;
        for (j=0; j<vdCurCalculatedMZ.size(); j++)
        {
            dErrorMZ = curPeak.dMZ - vdCurCalculatedMZ.at(j);
            if (fabs(dErrorMZ) <= this->pSettingInfo->dMassToleranceFragmentIons)
            {
                dCurPeakScore = calculatePeakFragmentMatch(curPeak, dCurrentPlausibilityScore, dErrorMZ);
                if (dCurPeakScore > dHighestPeakScore)
                {
                    dHighestPeakScore = dCurPeakScore;
                    iPeakRecordID = i;
                    iMassWindowID = j;
                }
            }
        }
    }
    
    return dHighestPeakScore;
}*/

double SpectrumUnit::calculateScoreFragmentAllPeaks_original(scoreInfo &scoreCompound,
                                                             const vector<double> &vdCalculatedMZ,
                                                             vector<double> &vdCurCalculatedMZ,
                                                             const fragType *pCurFragment, size_t &iPeakRecordID,
                                                             size_t &iMassWindowID, double dCurrentPlausibilityScore)
//scoreCompound: score information of the compound;
//vdCalculatedMZ: precalculated mz for all ion mass windows, if peak charge is NOT equal to this->iDefaultIonChargeState,
// it should be recalculated; vdCurCalculatedMZ: mz for all ion mass windows of the current fragment
//pCurFragment: pointer to the current fragment; iPeakRecordID: peak record id with the highest score;
//iMassWindowID: mass window id with the highest score;
//dCurrentPlausibilityScore: plausibility score of the current fragment; return the highest score
{
    double dHighestPeakScore = -1, dErrorMZ, dCurPeakScore;
    size_t i, j, iPeakNum;
    peakType curPeak;
    
    iPeakRecordID=-1;
    iMassWindowID=-1;
    
    iPeakNum = this->vPeaks.size();
    for (i = 0; i < iPeakNum; i++)
    {
        curPeak = this->vPeaks.at(i);
        if (curPeak.iPeakCharge != this->iDefaultIonChargeState)
            calculateAllPossibleMZ(vdCurCalculatedMZ, pCurFragment, curPeak.iPeakCharge);
        else
            vdCurCalculatedMZ = vdCalculatedMZ;
        for (j=0; j<vdCurCalculatedMZ.size(); j++)
        {
            dErrorMZ = curPeak.dMZ - vdCurCalculatedMZ.at(j);
            if (fabs(dErrorMZ) <= this->pSettingInfo->dMassToleranceFragmentIons)
            {
                dCurPeakScore = calculatePeakFragmentMatch(curPeak, dCurrentPlausibilityScore, dErrorMZ);
                if (dCurPeakScore > dHighestPeakScore)
                {
                    dHighestPeakScore = dCurPeakScore;
                    iPeakRecordID = i;
                    iMassWindowID = j;
                }
                if (dCurPeakScore > scoreCompound.vPeakFragInfo.at(i).dPeakScore)
                {
                    scoreCompound.vPeakFragInfo.at(i).dPeakScore = dCurPeakScore;
                    scoreCompound.vPeakFragInfo.at(i).sSmiles = pCurFragment->sSmiles;
                    scoreCompound.vPeakFragInfo.at(i).iFragmentationLevel = pCurFragment->iFragmentationLevel;
                    scoreCompound.vPeakFragInfo.at(i).iProtonOffset = this->viIonFragmentMassWindows.at(j);
                    scoreCompound.vPeakFragInfo.at(i).dMZError = curPeak.dMZ - vdCurCalculatedMZ.at(j);
                }
            }
        }
    }
    
    return dHighestPeakScore;
}

/*
bool SpectrumUnit::scoreFragment(scoreInfo &scoreCompound, fragType *pCurFragment, double dParentPlausibilityScore,
                                 double &dCurrentPlausibilityScore)
//scoreCompound: score information of the compound; pCurFragment: pointer to the current fragment
//dParentPlausibilityScore: plausibility score of the parent fragment
//dCurrentPlausibilityScore: palusibility score of the current fragment that is calculated in this function.
//return true if the current fragment is observed
{
    bool bDetected=false;
    vector<double> vdCalculatedMZ;
    vector<double> vdCurCalculatedMZ; //recalculate for peak charge is NOT equal to this->iDefaultIonChargeState
    double dHighestPeakScore = -1;
    //iPeakRecordID and iMassWindowID are for the current hightest score case
    size_t iPeakRecordID=-1, iMassWindowID=-1;
    
    dCurrentPlausibilityScore = calculatePlausibilityScore(pCurFragment, dParentPlausibilityScore);
    calculateAllPossibleMZ(vdCalculatedMZ, pCurFragment, this->iDefaultIonChargeState);
    dHighestPeakScore = calculateScoreFragmentAllPeaks(vdCalculatedMZ, vdCurCalculatedMZ, pCurFragment,
                                                       iPeakRecordID, iMassWindowID, dCurrentPlausibilityScore);
    
    if (dHighestPeakScore > 0)
    {
        bDetected = true;
        scoreCompound.dScore += dHighestPeakScore;
        if (scoreCompound.vPeakFragInfo.at(iPeakRecordID).dPeakScore < dHighestPeakScore)
        {
            scoreCompound.vPeakFragInfo.at(iPeakRecordID).dPeakScore = dHighestPeakScore;
            scoreCompound.vPeakFragInfo.at(iPeakRecordID).sSmiles = pCurFragment->sSmiles;
            scoreCompound.vPeakFragInfo.at(iPeakRecordID).iFragmentationLevel = pCurFragment->iFragmentationLevel;
            scoreCompound.vPeakFragInfo.at(iPeakRecordID).iProtonOffset = viIonFragmentMassWindows.at(iMassWindowID);
            scoreCompound.vPeakFragInfo.at(iPeakRecordID).dMZError = vPeaks.at(iPeakRecordID).dMZ
                                                                        - vdCurCalculatedMZ.at(iMassWindowID);
        }
    }
    
    return bDetected;
}*/

bool SpectrumUnit::scoreFragment_original(scoreInfo &scoreCompound, fragType *pCurFragment,
                                          double dParentPlausibilityScore, double &dCurrentPlausibilityScore)
//scoreCompound: score information of the compound; pCurFragment: pointer to the current fragment
//dParentPlausibilityScore: plausibility score of the parent fragment
//dCurrentPlausibilityScore: palusibility score of the current fragment that is calculated in this function.
//return true if the current fragment is observed
{
    bool bDetected=false;
    vector<double> vdCalculatedMZ;
    vector<double> vdCurCalculatedMZ; //recalculate for peak charge is NOT equal to this->iDefaultIonChargeState
    double dHighestPeakScore = -1;
    //iPeakRecordID and iMassWindowID are for the current hightest score case
    size_t iPeakRecordID=-1, iMassWindowID=-1;
    
    dCurrentPlausibilityScore = calculatePlausibilityScore(pCurFragment, dParentPlausibilityScore);
    calculateAllPossibleMZ(vdCalculatedMZ, pCurFragment, this->iDefaultIonChargeState);
    dHighestPeakScore = calculateScoreFragmentAllPeaks_original(scoreCompound,vdCalculatedMZ,vdCurCalculatedMZ,
                                                                pCurFragment,iPeakRecordID, iMassWindowID,
                                                                dCurrentPlausibilityScore);
    if (dHighestPeakScore > 0)
        bDetected = true;
    
    return bDetected;
}

void SpectrumUnit::buildBondInformation(fragType *pCurFragment, map<unsigned int, bondInfo> &mapBonds)
//pCurFragment: pointer to the current fragment
//pCurFragmentRWMol: pointer to the RWMol of current fragment; mapBonds: map of all bonds
{
    ROMol::EDGE_ITER firstB,lastB;
    bondInfo currentBond;
    
    mapBonds.clear();
    boost::tie(firstB,lastB) = pCurFragment->pFragmentRWMol->getEdges();
    while (firstB != lastB)
    {
        boost::shared_ptr<Bond> bond = (*(pCurFragment->pFragmentRWMol))[*firstB];
        currentBond = {};
        currentBond.iBondID = bond->getIdx();
        currentBond.iBeginAtomIdx = bond->getBeginAtomIdx();
        currentBond.iEndAtomIdx = bond->getEndAtomIdx();
        if (pSettingInfo->bGraphGrammar)
        {
            //check graph grammar
            assignBondWeightByGrammar(pCurFragment, currentBond);
        }else
        {
            currentBond.dBondWeight = 1.0;
        }
        mapBonds[currentBond.iBondID] = currentBond;
        ++firstB;
    }
}

void SpectrumUnit::assignBondWeightByGrammar(fragType *pCurFragment, bondInfo &currentBond)
//pCurFragment: pointer to the current fragment; currentBond: the current bond
{
    double dBondWeight = -10;
    size_t i;
    size_t iEdgeIdx;
    string sOriginalAnnotation = "";
    bool bPatternIdentified, bNoMatch = true;
    ParseRegularGraph ParseG(true);//bPartialParse = true
    
    updateEdgeAnnotation(pCurFragment->rgFragTopology, currentBond.iBondID, iEdgeIdx, sOriginalAnnotation);
    for (i = 0; i < pCurFragment->vbGrammarPattern.size(); i++)
        if (pCurFragment->vbGrammarPattern.at(i))
        {
            bPatternIdentified = ParseG.parseGraph(pCurFragment->rgFragTopology, *(pSettingInfo->vpSpecialGrammars.at(i)));
            if (bPatternIdentified)
            {
                bNoMatch = false;
                if (dBondWeight < pSettingInfo->vdGrammarWeights.at(i))
                    dBondWeight = pSettingInfo->vdGrammarWeights.at(i);
            }
        }
    if (bNoMatch)
        dBondWeight = 1.0;
    
    //recover update
    pCurFragment->rgFragTopology.vpSubgraphs.at(SUBGRAPHID)->vpEdges.at(iEdgeIdx)->sLabel = sOriginalAnnotation;
    
    currentBond.dBondWeight = dBondWeight;
}

void SpectrumUnit::updateEdgeAnnotation(RegularGraph &rgFragTopology, unsigned int iBondID, size_t &iEdgeIdx, string &sOriginalAnnotation)
//rgFragTopology: graph of the current fragment; iBondID: bond if of the current edge;
//iEdgeIdx: edge index of the vpEdges; sOriginalAnnotation: original string
{
    Edge *pCurrentEdge;
    size_t i;
    bool bEdgeIdentified = false;
    
    if (iBondID < rgFragTopology.vpSubgraphs.at(SUBGRAPHID)->vpEdges.size())
        pCurrentEdge = rgFragTopology.vpSubgraphs.at(SUBGRAPHID)->vpEdges.at(iBondID);
    else
        pCurrentEdge = nullptr;
    if ((pCurrentEdge != nullptr) && (pCurrentEdge->iEdgeId == iBondID))
    {
        iEdgeIdx = iBondID;
        bEdgeIdentified = true;
    }
    else
        for (i = 0; i < rgFragTopology.vpSubgraphs.at(SUBGRAPHID)->vpEdges.size(); i++)
        {
            pCurrentEdge = rgFragTopology.vpSubgraphs.at(SUBGRAPHID)->vpEdges.at(i);
            if (pCurrentEdge->iEdgeId == iBondID)
            {
                iEdgeIdx = i;
                bEdgeIdentified = true;
                break;
            }
        }
    if (bEdgeIdentified)
    {
        sOriginalAnnotation = pCurrentEdge->sLabel;
        pCurrentEdge->sLabel = "focus|"+sOriginalAnnotation;
    }else
    {
        cerr<<"cannot find edge with id "<<iBondID<<endl;
        exit(1);
    }
        
}

void SpectrumUnit::generateAllRingBondPairs(const vector<vector<int> > &viiRingBonds, vector<bondPair> &vRingBonds,
                                            set<unsigned int> &setRingBondsIdx,const map<unsigned int,bondInfo> &mapBonds)
//viiRingBonds: ring bond information; vRingBonds: vector of ring bond pairs;
//setRingBondsIdx: set of indexes of all ring bonds; mapBonds: map of all bonds
{
    bondPair currentBondPair;
    size_t i, j, k;
    
    setRingBondsIdx.clear();
    for (i = 0; i < viiRingBonds.size(); i++)
    {
        for (j = 0; j < (viiRingBonds.at(i).size() - 1); j++)
        {
            setRingBondsIdx.insert(viiRingBonds.at(i).at(j));
            for (k = (j + 1); k < viiRingBonds.at(i).size(); k++)
            {
                currentBondPair.ringBondOne = mapBonds.find(viiRingBonds.at(i).at(j))->second;
                currentBondPair.ringBondTwo = mapBonds.find(viiRingBonds.at(i).at(k))->second;
                vRingBonds.push_back(currentBondPair);
            }
        }
        setRingBondsIdx.insert(viiRingBonds.at(i).back());
    }
}

void SpectrumUnit::identifyAllLinearBonds(const RWMol *pCurFragmentRWMol, const set<unsigned int> &setRingBondsIdx,
                                          const map<unsigned int, bondInfo> &mapBonds, vector<bondInfo> &vLinearBonds)
//pCurFragmentRWMol: pointer to the RWMol of current fragment; setRingBondsIdx: set of indexes of all ring bonds;
//mapBonds: map of all bonds; vLinearBonds: vector of linear bonds;
{
    ROMol::EDGE_ITER firstB,lastB;
    unsigned int iCurrentBondId;
    
    boost::tie(firstB,lastB) = pCurFragmentRWMol->getEdges();
    while (firstB != lastB)
    {
        boost::shared_ptr<Bond> bond = (*pCurFragmentRWMol)[*firstB];
        iCurrentBondId = bond->getIdx();
        if (setRingBondsIdx.find(iCurrentBondId) == setRingBondsIdx.end())
            vLinearBonds.push_back(mapBonds.find(iCurrentBondId)->second);
        ++firstB;
    }
}

void SpectrumUnit::bondClassify(fragType *pCurFragment,vector<bondInfo> &vLinearBonds,vector<bondPair> &vRingBonds)
//pCurFragment: pointer to the current fragment; vLinearBonds: vector of linear bonds;
//vRingBonds: vector of ring bond pairs.
{
    vector<vector<int> > viiRingBonds;
    map<unsigned int, bondInfo> mapBonds;
    set<unsigned int> setRingBondsIdx;
    
    //The following two lines are moved to function completeNewFrag
    //if (!pCurFragment->pFragmentRWMol->getRingInfo()->isInitialized())
    //    MolOps::findSSSR(*(pCurFragment->pFragmentRWMol));
    viiRingBonds = pCurFragment->pFragmentRWMol->getRingInfo()->bondRings();
    buildBondInformation(pCurFragment, mapBonds);
    generateAllRingBondPairs(viiRingBonds, vRingBonds, setRingBondsIdx, mapBonds);
    identifyAllLinearBonds(pCurFragment->pFragmentRWMol, setRingBondsIdx, mapBonds, vLinearBonds);
}

void SpectrumUnit::completeNewFrag(fragType *pNewFrag, const fragType *pParentFragment,
                                   const vector<bondInfo> &vCurrentRemoveBonds, const bool &bParentFragmentDetected)
//pNewFrag: pointer to the new fragment; pParentFragment: point to the parent fragment
//vCurrentRemoveBonds: vector of removed bonds for generating new fragment
//bParentFragmentDetected: true, if the parent fragment is detected
{
    GML *pSubGraph;
    vector<bondInfo>::const_iterator const_itBond;
    
    if (!pNewFrag->pFragmentRWMol->getRingInfo()->isInitialized())
        MolOps::findSSSR(*(pNewFrag->pFragmentRWMol));
    
    pNewFrag->dFragmentMass = Descriptors::calcExactMW(*(pNewFrag->pFragmentRWMol));
    pNewFrag->molTopology = pNewFrag->pFragmentRWMol->getTopology();
    if (pSettingInfo->bGraphGrammar)
    {
        //pSubGraph = Database::MolGraph2GML(pNewFrag->molTopology);
        pSubGraph = Database::ROMol2GML(pNewFrag->pFragmentRWMol);
        pNewFrag->rgFragTopology.vpSubgraphs.push_back(pSubGraph);
    }
    pNewFrag->iFragmentationLevel = pParentFragment->iFragmentationLevel + 1;
    pNewFrag->bParentFragmentDetected = bParentFragmentDetected;
    pNewFrag->dBondWeight = 0;
    for (const_itBond = vCurrentRemoveBonds.begin(); const_itBond < vCurrentRemoveBonds.end(); ++const_itBond)
        pNewFrag->dBondWeight += const_itBond->dBondWeight;
    pNewFrag->sSmiles = MolToSmiles(*(pNewFrag->pFragmentRWMol));
    checkBondPatternsByGrammar(pNewFrag);
    
}

void SpectrumUnit::checkBondPatternsByGrammar(fragType *pCurrentFrag)
//pCurrentFrag: pointer to the current fragment
{
    bool bPatternIdentified;
    size_t i;
    ParseRegularGraph ParseG(true);//bPartialParse = true
    
    pCurrentFrag->vbGrammarPattern.clear();
    for (i = 0; i < pSettingInfo->vpGeneralGrammars.size(); i++)
    {
        bPatternIdentified = ParseG.parseGraph(pCurrentFrag->rgFragTopology, *(pSettingInfo->vpGeneralGrammars.at(i)));
        pCurrentFrag->vbGrammarPattern.push_back(bPatternIdentified);
    }
}

bool SpectrumUnit::nextTwoFragments(fragType *pCurFragment, const vector<bondInfo> &vCurrentRemoveBonds,
                                const bool &bCurrentFragmentDetected, fragType *pNewFragOne, fragType *pNewFragTwo)
//pCurFragment: pointer to the current fragment; vCurrentRemoveBonds: vector of bonds which should be removed
//bCurrentFragmentDetected: true if the current fragment is detected;
//pNewFragOne, pNewFragTwo: two new fragments generated by remove bond(s)
//return true if successfully generate two new fragments
{
    RWMol *pNewRWMol;
    bool bValidNewFragments = false;
    size_t i;
    vector<boost::shared_ptr<ROMol> > vROMol;
    
    pNewRWMol = new RWMol(*(pCurFragment->pFragmentRWMol));
    for (i = 0; i < vCurrentRemoveBonds.size(); i++)
        pNewRWMol->removeBond(vCurrentRemoveBonds.at(i).iBeginAtomIdx, vCurrentRemoveBonds.at(i).iEndAtomIdx);
    //cout<<MolToSmiles(*pNewRWMol)<<endl;
    vROMol = MolOps::getMolFrags(*(pNewRWMol), false);//!!!
    if (vROMol.size() == 2)
    {
        bValidNewFragments = true;
        pNewFragOne->pFragmentRWMol = new RWMol(*(vROMol.at(0).get()));
        completeNewFrag(pNewFragOne, pCurFragment, vCurrentRemoveBonds, bCurrentFragmentDetected);
        pNewFragTwo->pFragmentRWMol = new RWMol(*(vROMol.at(1).get()));
        completeNewFrag(pNewFragTwo, pCurFragment, vCurrentRemoveBonds, bCurrentFragmentDetected);
        //cout<<pNewFragOne->iFragmentationLevel<<endl;
    }
    delete pNewRWMol;
    return bValidNewFragments;
}

/*
void SpectrumUnit::scoreFollowingFragmentLevels(scoreInfo &scoreCompound, fragType *pCurFragment,
                                                const vector<bondInfo> &vLinearBonds,
                                                const vector<bondPair> &vRingBonds,
                                                const double &dCurrentPlausibilityScore,
                                                const bool & bCurrentFragmentDetected)
//scoreCompound: score information of the compound; pCurFragment: pointer to the current fragment
//vLinearBonds:vectors of all linear bonds;vRingBonds: ring bond pairs of the current fragment;
//dCurrentPlausibilityScore: plausibility score of the current fragment;
//bCurrentFragmentDetected: true, if the current fragment is detected
{
    vector<bondInfo> vCurrentRemoveBonds;
    size_t i, iLinearBondNum, iRingBondPairNum;
    fragType *pNextFragmentOne, *pNextFragmentTwo;
    bool bValidNewFragments;
    
    iLinearBondNum = vLinearBonds.size();
    iRingBondPairNum = vRingBonds.size();
    for (i = 0; i < iLinearBondNum; i++)
    {
        vCurrentRemoveBonds.clear();
        vCurrentRemoveBonds.push_back(vLinearBonds.at(i));
        pNextFragmentOne = new fragType;
        pNextFragmentTwo = new fragType;
        bValidNewFragments = nextTwoFragments(pCurFragment, vCurrentRemoveBonds, bCurrentFragmentDetected,
                                              pNextFragmentOne, pNextFragmentTwo);
        if (bValidNewFragments)
        {
            processFragment(scoreCompound, pNextFragmentOne, dCurrentPlausibilityScore);
            processFragment(scoreCompound, pNextFragmentTwo, dCurrentPlausibilityScore);
        }
        delete pNextFragmentOne->pFragmentRWMol;
        delete pNextFragmentOne;
        delete pNextFragmentTwo->pFragmentRWMol;
        delete pNextFragmentTwo;
    }
    if ((pSettingInfo->bBreakRings)&&(iRingBondPairNum > 0))
    {
        for (i = 0; i < iRingBondPairNum; i++)
        {
            vCurrentRemoveBonds.clear();
            vCurrentRemoveBonds.push_back(vRingBonds.at(i).ringBondOne);
            vCurrentRemoveBonds.push_back(vRingBonds.at(i).ringBondTwo);
            pNextFragmentOne = new fragType;
            pNextFragmentTwo = new fragType;
            bValidNewFragments = nextTwoFragments(pCurFragment, vCurrentRemoveBonds, bCurrentFragmentDetected,
                                                  pNextFragmentOne, pNextFragmentTwo);
            if (bValidNewFragments)
            {
                processFragment(scoreCompound, pNextFragmentOne, dCurrentPlausibilityScore);
                processFragment(scoreCompound, pNextFragmentTwo, dCurrentPlausibilityScore);
            }
            delete pNextFragmentOne->pFragmentRWMol;
            delete pNextFragmentOne;
            delete pNextFragmentTwo->pFragmentRWMol;
            delete pNextFragmentTwo;
        }
    }
}
*/

void SpectrumUnit::scoreFollowingFragmentLevels_original(scoreInfo &scoreCompound, fragType *pCurFragment,
                                                         const vector<bondInfo> &vLinearBonds,
                                                         const vector<bondPair> &vRingBonds,
                                                         const double &dCurrentPlausibilityScore,
                                                         const bool & bCurrentFragmentDetected)
//scoreCompound: score information of the compound; pCurFragment: pointer to the current fragment
//vLinearBonds:vectors of all linear bonds;vRingBonds: ring bond pairs of the current fragment;
//dCurrentPlausibilityScore: plausibility score of the current fragment;
//bCurrentFragmentDetected: true, if the current fragment is detected
{
    vector<bondInfo> vCurrentRemoveBonds;
    size_t i, iLinearBondNum, iRingBondPairNum;
    fragType *pNextFragmentOne, *pNextFragmentTwo;
    bool bValidNewFragments;
    
    iLinearBondNum = vLinearBonds.size();
    iRingBondPairNum = vRingBonds.size();
    for (i = 0; i < iLinearBondNum; i++)
    {
        vCurrentRemoveBonds.clear();
        vCurrentRemoveBonds.push_back(vLinearBonds.at(i));
        pNextFragmentOne = new fragType;
        pNextFragmentTwo = new fragType;
        bValidNewFragments = nextTwoFragments(pCurFragment, vCurrentRemoveBonds, bCurrentFragmentDetected,
                                              pNextFragmentOne, pNextFragmentTwo);
        if (bValidNewFragments)
        {
            processFragment_original(scoreCompound, pNextFragmentOne, dCurrentPlausibilityScore);
            processFragment_original(scoreCompound, pNextFragmentTwo, dCurrentPlausibilityScore);
        }
        delete pNextFragmentOne->pFragmentRWMol;
        delete pNextFragmentOne;
        delete pNextFragmentTwo->pFragmentRWMol;
        delete pNextFragmentTwo;
    }
    if ((pSettingInfo->bBreakRings)&&(iRingBondPairNum > 0))
    {
        for (i = 0; i < iRingBondPairNum; i++)
        {
            vCurrentRemoveBonds.clear();
            vCurrentRemoveBonds.push_back(vRingBonds.at(i).ringBondOne);
            vCurrentRemoveBonds.push_back(vRingBonds.at(i).ringBondTwo);
            pNextFragmentOne = new fragType;
            pNextFragmentTwo = new fragType;
            bValidNewFragments = nextTwoFragments(pCurFragment, vCurrentRemoveBonds, bCurrentFragmentDetected,
                                                  pNextFragmentOne, pNextFragmentTwo);
            if (bValidNewFragments)
            {
                processFragment_original(scoreCompound, pNextFragmentOne, dCurrentPlausibilityScore);
                processFragment_original(scoreCompound, pNextFragmentTwo, dCurrentPlausibilityScore);
            }
            delete pNextFragmentOne->pFragmentRWMol;
            delete pNextFragmentOne;
            delete pNextFragmentTwo->pFragmentRWMol;
            delete pNextFragmentTwo;
        }
    }
    
}

/*
void SpectrumUnit::processFragment(scoreInfo &scoreCompound, fragType *pCurFragment, double dParentPlausibilityScore)
//scoreCompound: score information of the compound; pCurFragment: pointer to the current fragment
//dParentPlausibilityScore: plausibility score of the parent fragment
{
    double dCurrentPlausibilityScore = 0;
    bool bCurrentFragmentDetected;
    vector<bondInfo> vLinearBonds;
    vector<bondPair> vRingBonds;
    
    // step 1: score current fragment (scoreFragment)
    bCurrentFragmentDetected=scoreFragment_original(scoreCompound,pCurFragment, dParentPlausibilityScore,
                                                    dCurrentPlausibilityScore);
    
    if (pCurFragment->iFragmentationLevel < pSettingInfo->iFragmentationDepth)
    {
        // step 2: generate list of linear bond candidates and list ofm_property ring bonds
        bondClassify(pCurFragment, vLinearBonds, vRingBonds);
        // step 3: in a loop, call ProcessFragment for scoring fragments
        scoreFollowingFragmentLevels(scoreCompound, pCurFragment, vLinearBonds, vRingBonds,
                                              dCurrentPlausibilityScore, bCurrentFragmentDetected);
    }
}*/

void SpectrumUnit::processFragment_original(scoreInfo &scoreCompound, fragType *pCurFragment,
                                            double dParentPlausibilityScore)
//scoreCompound: score information of the compound; pCurFragment: pointer to the current fragment
//dParentPlausibilityScore: plausibility score of the parent fragment
{
    double dCurrentPlausibilityScore = 0;
    bool bCurrentFragmentDetected;
    vector<bondInfo> vLinearBonds;
    vector<bondPair> vRingBonds;
    
    // step 1: score current fragment (scoreFragment)
    bCurrentFragmentDetected=scoreFragment_original(scoreCompound,pCurFragment, dParentPlausibilityScore,
                                                    dCurrentPlausibilityScore);
    
    if (pCurFragment->iFragmentationLevel < pSettingInfo->iFragmentationDepth)
    {
        // step 2: generate list of linear bond candidates and list ofm_property ring bonds
        bondClassify(pCurFragment, vLinearBonds, vRingBonds);
        // step 3: in a loop, call ProcessFragment for scoring fragments
        scoreFollowingFragmentLevels_original(scoreCompound, pCurFragment, vLinearBonds, vRingBonds,
                                              dCurrentPlausibilityScore, bCurrentFragmentDetected);
    }
}

void SpectrumUnit::ProcessOneCompound(const vector<dbRecordType> & vCompoundDB, int iCurrentCompoundIdx)
//pCurrentCompound: pointer to the current compound
{
    scoreInfo currentCompoundScore;
    
    currentCompoundScore = scoreCompound(vCompoundDB.at(iCurrentCompoundIdx).pCompound);
    currentCompoundScore.iCompoundIdx = iCurrentCompoundIdx;
    this->vScoreMatches.push_back(currentCompoundScore);
}

void SpectrumUnit::Process(const vector<dbRecordType> & vCompoundDB)
//vCompoundDB: vector of all compounds in the compound database
{
    int i;
    getRelatedCompoundsFromDB(vCompoundDB);
    if ((iLowerIndexDB >= 0) && (iUpperIndexDB >= 0))
    {
        PreprocessPeaks();
        //use openmp
        //#pragma omp parallel for \
        //shared(vCompoundDB) private(i) \
        //schedule(guided)
        for (i = iLowerIndexDB; i <= iUpperIndexDB; i++)
        {
            //cout<<"processing "<< (i - iLowerIndexDB + 1)<< "out of "<<(iUpperIndexDB - iLowerIndexDB + 1)<<endl;
            ProcessOneCompound(vCompoundDB, i);
        }
        sort(this->vScoreMatches.begin(), this->vScoreMatches.end(), higherCompoundScore);
    }
}

void SpectrumUnit::OutputResults_internal()
{
    string sOutput = "", sOutputFileName;
    size_t i, j, iCurrentCompoundRank = 0, iRealCompoundRank = 0;
    ofstream OutputFileStream;
    double dOldScore = 0;
    vector<scoreInfo>::const_iterator itScoreInfo;
    vector<peakFragmentInfo>::const_iterator itPeakScoreInfo;
    
    sOutputFileName = pSettingInfo->sOutputDirectory + this->sIdentifier + ".output.txt";
    OutputFileStream.open(sOutputFileName, ios::out);
    i = 0;
    for (itScoreInfo = vScoreMatches.begin(); itScoreInfo != vScoreMatches.end(); ++itScoreInfo)
    {
        if (i==0)
            iCurrentCompoundRank = 1;
        else
            if (dOldScore > itScoreInfo->dScore)
                iCurrentCompoundRank = i+1;
        if (this->sInchi == itScoreInfo->sInchi)
            iRealCompoundRank = iCurrentCompoundRank;
        dOldScore = itScoreInfo->dScore;
        sOutput+=to_string(iCurrentCompoundRank)+"\t"+to_string(itScoreInfo->dScore)+"\tNA\t"
                    +itScoreInfo->sCompoundIdentifier+"\t"+itScoreInfo->sInchi+"\tNA\t"+itScoreInfo->sExplainedPeaks+"\n";
        sOutput+="mz\tnormalized_int\tpeak_score\tsmiles\tfragment_level\tproton_offset\tmz_error\n";
        j=0;
        for (itPeakScoreInfo=itScoreInfo->vPeakFragInfo.begin();itPeakScoreInfo!=itScoreInfo->vPeakFragInfo.end();
             ++itPeakScoreInfo)
        {
            if (itPeakScoreInfo->dPeakScore > 0)
                sOutput+=to_string(this->vPeaks.at(j).dMZ)+"\t"+to_string(this->vPeaks.at(j).dIntensity)+"\t"
                        +to_string(itPeakScoreInfo->dPeakScore)+"\t"+itPeakScoreInfo->sSmiles+"\t"
                        +to_string(itPeakScoreInfo->iFragmentationLevel)+"\t"+to_string(itPeakScoreInfo->iProtonOffset)
                        +"\t"+to_string(itPeakScoreInfo->dMZError)+"\n";
            j++;
        }
        i++;
    }
    sOutput = ">" + to_string(iRealCompoundRank) + "\t" + this->sIdentifier + "\n" + sOutput;
    if (OutputFileStream.is_open())
    {
        OutputFileStream<<sOutput;
    }else
    {
        cerr<<"cannot open "<<sOutputFileName<<endl;
        exit(1);
    }
    OutputFileStream.close();
}

bool SpectrumUnit::higherCompoundScore(scoreInfo score1, scoreInfo score2)
//return true if score1 has higher score
//score1 and score2 are score information of two compounds against the current spectrum
{
    return (score1.dScore > score2.dScore);
}

void SpectrumUnit::getRelatedCompoundsFromDB(const vector<dbRecordType> & vCompoundDB)
//vCompoundDB: vector of all compounds in the compound database
{
    double dPrecursorMass, dMinPrecursorMass, dMaxPrecursorMass;
    
    if (pSettingInfo->bPositiveDefaultPolarity)
        dPrecursorMass = this->dPrecursorMZ * this->iPrecursorCharge - this->iPrecursorCharge * pSettingInfo->dProtonMass;
    else
        dPrecursorMass = (-1)*(this->dPrecursorMZ * this->iPrecursorCharge + this->iPrecursorCharge * pSettingInfo->dProtonMass);
    this->dPrecursorMass = dPrecursorMass;
    dMinPrecursorMass = dPrecursorMass - pSettingInfo->dMassToleranceParentIon;
    dMaxPrecursorMass = dPrecursorMass + pSettingInfo->dMassToleranceParentIon;
    iLowerIndexDB = locateBinarySearch_LowerBound(dMinPrecursorMass, vCompoundDB);
    iUpperIndexDB = locateBinarySearch_UpperBound(dMaxPrecursorMass, vCompoundDB);
    
    if ((iLowerIndexDB == -1) || (iUpperIndexDB == -1) || (iLowerIndexDB > iUpperIndexDB))
    {
        iLowerIndexDB = -1;
        iUpperIndexDB = -1;
    }
    
}

int SpectrumUnit::locateBinarySearch_LowerBound(double dMinPrecursorMass, const vector<dbRecordType> & vCompoundDB)
//dMinPrecursorMass: minimum acceptable mass of precursor; vCompoundDB: vector of all compounds in the compound database
//return the index of lightest compound from DB within the mass range
{
    int iUpperIndex, iLowerIndex, iMiddleIndex;
    
    iUpperIndex = (int)(vCompoundDB.size() - 1);
    if (vCompoundDB.at(iUpperIndex).dMass < dMinPrecursorMass)
        return -1;
    iLowerIndex = 0;
    if (vCompoundDB.at(iLowerIndex).dMass >= dMinPrecursorMass) //!!!
        return iLowerIndex;
    while ((iUpperIndex - iLowerIndex) > 1)
    {
        iMiddleIndex = (iLowerIndex + iUpperIndex) / 2;
        if (vCompoundDB.at(iMiddleIndex).dMass >= dMinPrecursorMass)
            iUpperIndex = iMiddleIndex;
        else
            iLowerIndex = iMiddleIndex;
    }
    
    return iUpperIndex;
}

int SpectrumUnit::locateBinarySearch_UpperBound(double dMaxPrecursorMass, const vector<dbRecordType> & vCompoundDB)
//dMaxPrecursorMass: maximum acceptable mass of precursor; vCompoundDB: vector of all compounds in the compound database
//return the index of heaviest compound from DB within the mass range
{
    int iUpperIndex, iLowerIndex, iMiddleIndex;
    
    iUpperIndex = (int)(vCompoundDB.size() - 1);
    if (vCompoundDB.at(iUpperIndex).dMass <= dMaxPrecursorMass) //!!!
        return iUpperIndex;
    iLowerIndex = 0;
    if (vCompoundDB.at(iLowerIndex).dMass > dMaxPrecursorMass)
        return -1;
    while ((iUpperIndex - iLowerIndex) > 1)
    {
        iMiddleIndex = (iLowerIndex + iUpperIndex) / 2;
        if (vCompoundDB.at(iMiddleIndex).dMass > dMaxPrecursorMass)
            iUpperIndex = iMiddleIndex;
        else
            iLowerIndex = iMiddleIndex;
    }
    
    return iLowerIndex;
}




