//
//  SpectrumUnit.h
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/20/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#ifndef __midas_cpp__SpectrumUnit__
#define __midas_cpp__SpectrumUnit__

#include <stdio.h>
#include <string>
#include <vector>

#include <GraphMol/SmilesParse/SmilesWrite.h>

#include "Database.h"
#include "SettingInfo.h"

#define SUBGRAPHID 0

using namespace std;

struct peakType
{
    double dMZ; //m/z
    double dIntensity; //Intensity
    double dResolution; //Resolution
    double dBaseline; //Baseline
    double dNoise; // Noise
    int    iPeakCharge; //Charge
};

struct peakFragmentInfo
{
    string sSmiles = ""; // SMILES format of the fragment
    int iProtonOffset; // proton offset
    double dMZError; //m/zError;dMeasured - dCalculated;dMeasuredMZ - ((current_dMass+current_mz_offset*dHMass)/current_z)
    int iFragmentationLevel; // fragmentation level in the fragmentation tree
    double dPeakScore = -100; // score of this peak fragment match
};

struct scoreInfo
{
    //int iRank; // rank starting from 1
    double dParentMassError; //dCompoundMass - dSpectrumPrecursorMass
    string sExplainedPeaks; // NumOfExplainedPeaks>TotalNumOfPeaks
    double dScore; //score of the spectrum compound match
    string sModifications = "NA"; // the current version doesn't consider modification
    vector<peakFragmentInfo> vPeakFragInfo; //peak score information
    size_t iCompoundIdx; //compound index in database
    string sCompoundIdentifier; // compound Identifier
    string sInchi; //Inchi of compound
    string sCompoundName; //Compound Name
    string sCompoundLinks; //Compound Links
};

struct fragType
{
    double dFragmentMass; //mass of fragment
    RWMol* pFragmentRWMol = nullptr; // pointer referring to the mol of fragment
    MolGraph molTopology; //MolGraph Topology of the fragment
    RegularGraph rgFragTopology; //graph of fragment topology
    int iFragmentationLevel; // fragmentation level in the fragmentation tree
    bool bParentFragmentDetected; //true if parent fragment is detected; true for intact compound (root of fragment tree)
    double dBondWeight;// by default, breaking 1 linear bond = 1, breaking two ring bonds = 2
    string sSmiles; // SMILES format of the fragment
    vector<bool> vbGrammarPattern; //element is true if the corresponding general bond pattern is identified in the fragment
};

struct bondInfo
{
    unsigned int iBondID;
    unsigned int iBeginAtomIdx;
    unsigned int iEndAtomIdx;
    double dBondWeight;
};

struct bondPair
{
    bondInfo ringBondOne;
    bondInfo ringBondTwo;
};

class SpectrumUnit
{
    const SettingInfo *pSettingInfo; //point to setting
    int iLowerIndexDB; //index lightest compound within the range
    int iUpperIndexDB; //index of heaviest compound within the range
    double dHalf_mass_tolerance_fragment_ions; // half of Fragment mass tolerance in Da
    double dPrecursorMass; //calculated mass of precursor
    vector<int> viIonFragmentMassWindows; // fragment ion mass window
    int iDefaultIonChargeState = 1; //default ion (or peak) charge state, also used in ParseSpectraFile
    
    //locate the range of compounds
    void getRelatedCompoundsFromDB(const vector<dbRecordType> & vCompoundDB);
    //locate the index of lightest compound from DB within the mass range
    int locateBinarySearch_LowerBound(double dMinPrecursorMass, const vector<dbRecordType> & vCompoundDB);
    //locate the index of heaviest compound from DB within the mass range
    int locateBinarySearch_UpperBound(double dMaxPrecursorMass, const vector<dbRecordType> & vCompoundDB);
    //preprocess peaks
    void PreprocessPeaks();
    //return true if mz of record_i is less than that of record_j.
    static bool lessMZ(peakType record_i, peakType record_j);
    void NormalizeIntensity(); //normalize intensities of peaks
    // calculate pnorm
    double mypnorm(double dMean, double dStandardDeviation, double dRandomVariable);
    // calculate Lamda (Error Score)
    double calLamda(double dErrorMZ);
    
    scoreInfo scoreCompound(const compoundType * pCompound);//score current compound against the current spectrum
    //process current fragment
    /*void processFragment(scoreInfo &scoreCompound, fragType *pCurFragment, double dParentPlausibilityScore);*/
    //process current fragment, midas_1.0 version
    void processFragment_original(scoreInfo &scoreCompound, fragType *pCurFragment, double dParentPlausibilityScore);
    //calculate score of the fragment
    /*bool scoreFragment(scoreInfo &scoreCompound, fragType *pCurFragment, double dParentPlausibilityScore,
                       double &dCurrentPlausibilityScore);*/
    //calculate score of the fragment, midas_1.0 version
    bool scoreFragment_original(scoreInfo &scoreCompound, fragType *pCurFragment, double dParentPlausibilityScore,
                                double &dCurrentPlausibilityScore);
    //calculate the current Plausibility Score
    double calculatePlausibilityScore(const fragType *pCurFragment, const double dParentPlausibilityScore);
    //calculate all possible mz of the current fragment
    void calculateAllPossibleMZ(vector<double> & vdPossibleMZ, const fragType *pCurFragment, int iCurrentCharge);
    //calculate the score between a specific peak and a specific fragment with a specific proton offset.
    double calculatePeakFragmentMatch(const peakType &curPeak, double dCurrentPlausibilityScore, double dErrorMZ);
    //calculate the highest score of a fragment
    /*double calculateScoreFragmentAllPeaks(const vector<double> & vdCalculatedMZ, vector<double> &vdCurCalculatedMZ,
                                          const fragType *pCurFragment, size_t &iPeakRecordID, size_t &iMassWindowID,
                                          double dCurrentPlausibilityScore);*/
    //calculate the highest score of a fragment, midas_1.0 version
    double calculateScoreFragmentAllPeaks_original(scoreInfo &scoreCompound, const vector<double> & vdCalculatedMZ,
                                                   vector<double> &vdCurCalculatedMZ, const fragType *pCurFragment,
                                                   size_t &iPeakRecordID, size_t &iMassWindowID,
                                                   double dCurrentPlausibilityScore);

    //assign bond weight and classify bonds into linear bonds and ring bonds
    void bondClassify(fragType *pCurFragment, vector<bondInfo> &vLinearBonds, vector<bondPair> &vRingBonds);
    //store all bond information
    void buildBondInformation(fragType *pCurFragment, map<unsigned int, bondInfo> &mapBonds);
    //organize ring bond pairs
    void generateAllRingBondPairs(const vector<vector<int> > &viiRingBonds, vector<bondPair> &vRingBonds,
                                  set<unsigned int> &setRingBondsIdx, const map<unsigned int,bondInfo> &mapBonds);
    //identify all linear bonds
    void identifyAllLinearBonds(const RWMol *pCurFragmentRWMol, const set<unsigned int> &setRingBondsIdx,
                                const map<unsigned int, bondInfo> &mapBonds, vector<bondInfo> &vLinearBonds);
    //check whether specific bond patterns are in the given fragment
    void checkBondPatternsByGrammar(fragType *pCurrentFrag);
    //score fragements in the following fragment levels
    /*void scoreFollowingFragmentLevels(scoreInfo &scoreCompound, fragType *pCurFragment,
                                      const vector<bondInfo> &vLinearBonds, const vector<bondPair> &vRingBonds,
                                      const double &dCurrentPlausibilityScore, const bool & bCurrentFragmentDetected);*/
    //score fragements in the following fragment levels
    void scoreFollowingFragmentLevels_original(scoreInfo &scoreCompound, fragType *pCurFragment,
                                               const vector<bondInfo> &vLinearBonds, const vector<bondPair> &vRingBonds,
                                               const double &dCurrentPlausibilityScore,
                                               const bool & bCurrentFragmentDetected);
    //generate next two fragments
    bool nextTwoFragments(fragType *pCurFragment, const vector<bondInfo> &vCurrentRemoveBonds,
                                const bool &bCurrentFragmentDetected, fragType *pNewFragOne, fragType *pNewFragTwo);
    //complete all fields of the new fragment
    void completeNewFrag(fragType *pNewFrag, const fragType *pParentFragment, const vector<bondInfo> &vCurrentRemoveBonds,
                         const bool &bParentFragmentDetected);
    //process one compound
    void ProcessOneCompound(const vector<dbRecordType> & vCompoundDB, int iCurrentCompoundIdx);
    //assign bond weight using graph grammars
    void assignBondWeightByGrammar(fragType *pCurFragment, bondInfo &currentBond);
    //update annotation of the current edge
    void updateEdgeAnnotation(RegularGraph &rgFragTopology, unsigned int iBondID, size_t &iEdgeIdx, string &sOriginalAnnotation);
    
public:
    double dPrecursorMZ; //Precursor M/Z
    string sInchi; //Inchi of the real compound associated with the spectrum (not available under official mode)
    string sScanNumber; //scan number of the spectrum
    string sFileName; //file name of the spectra file
    string sRetentionTime; //retention time
    string sScanType; //spectrum type: PositiveIon or NegativeIon
    string sIdentifier; //identifier of the spectrum
    int    iPrecursorCharge; //precursor charge
    vector<peakType> vPeaks; //vector of peaks
    vector<scoreInfo> vScoreMatches; //scores of matched compounds
    
    SpectrumUnit(const string &sInputFileName, const string &sScanType, int iDefaultPrecursorChargeState,
                 const SettingInfo * pSettingInfo);
    ~SpectrumUnit();
    
    //Process the spectrum
    void Process(const vector<dbRecordType> & vCompoundDB);
    void OutputResults_internal(); //Output results under internal mode 
    
    static bool higherCompoundScore(scoreInfo score1, scoreInfo score2);// return true if score1 has higher score

};

#endif /* defined(__midas_cpp__SpectrumUnit__) */
